@@ -1,6 +1,5 @@
 #############################################################################
 #
-#	$Id: protocol.py,v 2.94.2.40 2010/05/22 16:31:10 irmen Exp $
 #	Pyro Protocol Adapters
 #
 #	This is part of "Pyro" - Python Remote Objects
@@ -8,6 +7,7 @@
 #
 #############################################################################
 
+from __future__ import with_statement
 import socket, struct, os, time, sys, hmac, types, random, errno, select
 import imp, marshal, new, __builtin__
 try:
@@ -175,6 +175,26 @@ def set_sock_keepalive(sock):
 		except:
 			Pyro.config.PYRO_SOCK_KEEPALIVE=0    # it didn't work--disable keepalives.
 
+# set socket to not inherit in subprocess
+try:
+	import fcntl
+	def set_sock_no_inherit(sock):
+		# Mark the given socket fd as non-inheritable (posix)
+		fd = sock.fileno() 
+		flags = fcntl.fcntl(fd, fcntl.F_GETFD) 
+		fcntl.fcntl(fd, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)
+except ImportError:
+	# no fcntl available, try the windows version
+	try: 
+		from ctypes import windll, WinError 
+		def set_sock_no_inherit(sock): 
+			# mark the given socket fd as non-inheritable (Windows).
+			if not windll.kernel32.SetHandleInformation(sock.fileno(), 1, 0): 
+				raise WinError() 	
+	except ImportError:
+		# nothing available, define a dummy function
+		def set_sock_no_inherit(sock):
+			pass
 
 #------ PYRO: adapter (default Pyro wire protocol)
 #------ This adapter is for protocol version 4 ONLY
@@ -237,8 +257,7 @@ class PYROAdapter(object):
 		if URI.protocol not in ('PYRO', 'PYROLOC'):
 			Log.error('PYROAdapter','incompatible protocol in URI:',URI.protocol)
 			raise ProtocolError('incompatible protocol in URI')
-		try:
-			self.lock.acquire()   # only 1 thread at a time can bind the URI
+		with self.lock:   # only 1 thread at a time can bind the URI
 			try:
 				self.URI=URI
 				sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
@@ -269,8 +288,6 @@ class PYROAdapter(object):
 			except socket.error:
 				Log.msg('PYROAdapter','connection failed to URI',str(URI))
 				raise ProtocolError('connection failed')
-		finally:
-			self.lock.release()
 
 	def resolvePYROLOC_URI(self, newProtocol):
 		# This method looks up the object URI referenced by
@@ -433,19 +450,14 @@ class PYROAdapter(object):
 		return answer
 
 	def remoteInvocation(self, method, flags, *args):
-		try:
-			self.lock.acquire() # only 1 thread at a time may use this connection to call a remote method
+		with self.lock:
+			# only 1 thread at a time may use this connection to call a remote method
 			return self._remoteInvocation(method, flags, *args)
-		finally:
-			self.lock.release()
 
 	def _remoteInvocation(self, method, flags, *args):
 		if 'conn' not in self.__dict__.keys():
-			Log.msg('PYROAdapter','no connection, trying to bind again')
-			if 'URI' in self.__dict__.keys():
-				self.bindToURI(self.URI)
-			else:
-				raise ProtocolError('trying to rebind, but was never bound before')
+			Log.error('PYROAdapter','no connection available in remoteinvocation')
+			raise ProtocolError('no connection available in remoteinvocation')
 		if method in self.onewayMethods:
 			flags |= Pyro.constants.RIF_Oneway
 		body=pickle.dumps((self.URI.objectID,method,flags,args),Pyro.config.PYRO_PICKLE_FORMAT)
@@ -458,7 +470,6 @@ class PYROAdapter(object):
 			# may be catching the keyboardinterrupt in their code. We should probably be on the
 			# safe side and release the proxy connection in this case too, because they might
 			# be reusing the proxy object after catching the exception...
-			sys.stderr.write("*********************** ERROR IN SOCK SEND *************") #XXX
 			self.release()
 			raise
 		else:
@@ -634,19 +645,24 @@ class PYROAdapter(object):
 		#  (object ID, method name, flags, (arg1,arg2,...))
 		importer=fromlist=None
 		try:
-			try:
+			if Pyro.config.PYRO_MOBILE_CODE:
 				# install a custom importer to intercept any extra needed modules
 				# when unpickling the request just obtained from the client
-				imp.acquire_lock()
-				importer=agent_import(__builtin__.__import__)
-				__builtin__.__import__=importer
+				try:
+					imp.acquire_lock()
+					importer=agent_import(__builtin__.__import__)
+					__builtin__.__import__=importer
+					req=self._unpickleRequest(pflags, body)
+				finally:
+					__builtin__.__import__=importer.orig_import
+					imp.release_lock()
+			else:
+				# no mobile code; just unpickle the stuff without a custom importer.
 				req=self._unpickleRequest(pflags, body)
-				if type(req)!=tuple or len(req)!=4 or type(req[3])!=tuple:
-					# sanity check failed
-					raise ProtocolError("invalid request data format")
-			finally:
-				__builtin__.__import__=importer.orig_import
-				imp.release_lock()
+
+			if type(req)!=tuple or len(req)!=4 or type(req[3])!=tuple:
+				# sanity check failed
+				raise ProtocolError("invalid request data format")
 
 		except ImportError,x:
 			if Pyro.config.PYRO_MOBILE_CODE:
@@ -830,8 +846,7 @@ class PYROSSLAdapter(PYROAdapter):
 		if URI.protocol not in ('PYROSSL','PYROLOCSSL'):
 			Log.error('PYROSSLAdapter','incompatible protocol in URI:',URI.protocol)
 			raise ProtocolError('incompatible protocol in URI')
-		try:
-			self.lock.acquire()   # only 1 thread at a time can bind the URI
+		with self.lock:   # only 1 thread at a time can bind the URI
 			try:
 				self.URI=URI
 				sock = SSL.Connection(self.ctx,socket.socket(socket.AF_INET, socket.SOCK_STREAM))
@@ -857,8 +872,6 @@ class PYROSSLAdapter(PYROAdapter):
 			except socket.error:
 				Log.msg('PYROSSLAdapter','connection failed to URI',str(URI))
 				raise ProtocolError('connection failed')
-		finally:
-			self.lock.release()
 
 	def _sendConnect(self, sock, ident):
 		return PYROAdapter._sendConnect(self, sock, ident)
@@ -1008,6 +1021,7 @@ class TCPServer(object):
 			# create server socket for new connections
 			self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
 			set_reuse_addr(self.sock)
+			set_sock_no_inherit(self.sock)
 			self.sock.bind((host,port))
 			self.sock.listen(Pyro.config.PYRO_TCP_LISTEN_BACKLOG)
 			if self._ssl_server:
